from __future__ import annotations

from enum import Enum

from . import device as device

class Type(Enum):
    @property
    def Void(self) -> int: ...
    @property
    def Bool(self) -> int: ...
    @property
    def Uint16(self) -> int: ...
    @property
    def Int16(self) -> int: ...
    @property
    def Uint32(self) -> int: ...
    @property
    def Int32(self) -> int: ...
    @property
    def Uint64(self) -> int: ...
    @property
    def Int64(self) -> int: ...
    @property
    def Float(self) -> int: ...
    @property
    def Double(self) -> int: ...
    @property
    def ComplexFloat(self) -> int: ...
    @property
    def ComplexDouble(self) -> int: ...
