#ifndef CYTNX_CORE_H_
#define CYTNX_CORE_H_

#include <iostream>
#include <typeinfo>

// error entry header
#include <cytnx_core/errors/cytnx_error.hpp>

#include <cytnx_core/Device.hpp>
#include <cytnx_core/Type.hpp>

#endif  // CYTNX_CORE_H_
